from collections import defaultdict


class Event(list):
    def __call__(self, *args, **kwargs):
        for f in self:
            f(*args, **kwargs)


def create_events(*names):
    return type("EventSystem", (), {name: Event() for name in names})



import re
import pandas as pd
import operator
import inspect
from pathlib import Path
from collections import defaultdict
from functools import partial
from torch.utils.tensorboard import SummaryWriter

from .utils import EMAMeter


class Logger(SummaryWriter):
    def __init__(self, log_dir, smoothing=[], postfix="", **kwargs):
        """
        Args:
            *args, **kwargs
            postfix: postfix appended to the tag when added to tensorboard.
        """
        super().__init__(log_dir, **kwargs)
        postfix = re.sub("/+", "/", str(postfix).strip("/"))
        if postfix:
            postfix = "/" + postfix
        self._smoothing = smoothing
        self._postfix = postfix
        self._meters = defaultdict(EMAMeter)
        self._buffer = defaultdict(list)
        self._delayed_addings = []
        self._make_delayed_adders()

    def _make_delayed_adders(self):
        """
        Make the adder functions lazy and only to complete in the render function.
        If the adder is add_scalar or add_text, aslo save it to buffer for rendering.
        """
        for name, method in inspect.getmembers(self, inspect.ismethod):
            parameters = inspect.signature(method).parameters
            if name.startswith("add_") and "global_step" in parameters:

                def delayed(tag, value, _name=name, _method=method, **kwargs):
                    curried = partial(_method, tag + self._postfix, value, **kwargs)
                    self._delayed_addings.append(curried)
                    if _name in ["add_scalar", "add_text"]:
                        self._buffer[tag].append(value)

                setattr(self, name, delayed)

    def _priortize(self, l, priority):
        def helper(pair):
            i, s = pair
            p = len(l) - i
            for j, pattern in enumerate(priority):
                if re.match(pattern, s) is not None:
                    p = len(l) + len(priority) - j
            return "\0x00" * p + s

        return map(operator.itemgetter(1), sorted(enumerate(l), key=helper))

    @staticmethod
    def _prettify(value):
        try:
            if int(value) != value:
                value = f"{value:.4g}"
        except:
            pass
        return value

    def _render_value(self, tag):
        value = self._buffer[tag][-1]
        for pattern in self._smoothing:
            if re.match(pattern, tag) is not None:
                value = self._meters[tag](value)
        return self._prettify(value)

    def _perform_addings(self, global_step):
        for adding in self._delayed_addings:
            adding(global_step=global_step)
        self.flush()
        self._delayed_addings.clear()

    def render(self, global_step, priority=[]):
        self._perform_addings(global_step)
        self._buffer["iteration"].append(global_step)
        priority = ["iteration"] + priority
        tags = self._priortize(self._buffer, priority)
        lines = [f"{tag}: {self._render_value(tag)}" for tag in tags]
        return lines

    def keys(self):
        yield from self._buffer.keys()

    def values(self):
        yield from self._buffer.values()

    def items(self):
        yield from self._buffer.items()

    def average(self, tag):
        values = self._buffer[tag]
        return sum(values) / len(values)

    def __iter__(self):
        yield from self._buffer



import argparse
import numpy as np
import torch
import operator
import zouqi

from .base import BaseRunner


class LegacyRunner(BaseRunner):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def feed(self, x):
        return self.model(x)

    def criterion(self, x, y):
        raise NotImplementedError

    def prepare_batch(self, batch):
        raise NotImplementedError

    def step(self, batch):
        args = self.args

        model = self.model
        logger = self.logger

        x, y = self.prepare_batch(batch)

        if self.training:
            optimizer = self.optimizer

            optimizer.set_lr(self.args.lr())

            x = self.feed(x)
            loss = self.criterion(x, y)

            if self.model.amp is None:
                (loss / args.update_every).backward()
            else:
                with model.amp.scale_loss(loss / args.update_every, optimizer) as loss:
                    loss.backward()

            if (model.iteration + 1) % args.update_every == 0:
                optimizer.step()
                optimizer.zero_grad()

            logger.add_scalar("lr", optimizer.get_lr())
        else:
            with torch.no_grad():
                x = self.feed(x)
                loss = self.criterion(x, y)

        logger.add_scalar("loss", loss.item())

    @zouqi.command(inherit=True)
    def train(self, *args, update_every: int = 1, **kwargs):
        self.update_args(dict(update_every=update_every))
        super().train(*args, **kwargs)



import os
import tqdm
import time
import shutil
import argparse
import inspect
import numpy as np
import torch
import torch.nn as nn
import functools
from collections import defaultdict
from torch.utils.data import DataLoader
from pathlib import Path

import zouqi

from torchzq.parsing import boolean, flag, choices
from torchzq.saver import Saver
from torchzq.logging import Logger
from torchzq.utils import Timer
from torchzq.scheduler import Scheduler
from torchzq.events import create_events


def parse_modes(cls):
    modes = []
    for base in cls.__bases__:
        modes += parse_modes(base)
    modes += vars(cls).get("modes", [])
    return list(set(modes))


class MetaRunner(type):
    """
    Runner can be running in different modes, e.g. train, validate, test etc.
    Different mode will leads to different function behavior.
    This meta class automatically parse modes from the mode-determined function call.
    The mode-determined function decides what would be the mode it runs on and locks
    the mode variable until it returns.
    """

    def __new__(mcls, name, bases, attrs):
        @property
        def mode(self):
            try:
                return self._mode
            except:
                raise AttributeError(
                    "Mode is not set, please call from a mode function to start."
                )

        attrs["mode"] = mode

        cls = super().__new__(mcls, name, bases, attrs)

        modes = parse_modes(cls)

        for mode in modes:
            try:
                f = getattr(cls, mode)
            except:
                raise AttributeError(f'mode "{mode}" is not defined in your class.')

            def wrap(f):
                @functools.wraps(f)
                def wrapped(self, *args, **kwargs):
                    # change the current mode to the called function
                    if not getattr(self, "_mode_locked", False):
                        self._mode = f.__name__
                        self._mode_locked = True
                    ret = f(self, *args, **kwargs)
                    self._mode_locked = False
                    return ret

                return wrapped

            setattr(cls, mode, wrap(f))

        return cls


class BaseRunner(metaclass=MetaRunner):
    # different modes that a command could be in
    modes = ["train", "validate"]

    saver = None
    scheduler = None
    model = None
    optimizer = None

    def __init__(
        self,
        name: str = "Unnamed",
        batch_size: int = 32,
        nj: int = min(os.cpu_count(), 12),
        device: str = "cuda",
        ckpt_dir: Path = "ckpt",
        strict_loading: boolean = True,
        log_dir: Path = "logs",
        quiet: flag = False,
        split: str = None,
        amp_level: choices(["O0", "O1", "O2", "O3"]) = None,
    ):
        self.update_args(locals(), "self")
        self.events = create_events(
            "epoch_started",
            "epoch_completed",
            "iteration_started",
            "iteration_completed",
        )

    @property
    def name(self):
        return self.args.name

    @property
    def training(self):
        return self.mode == "train"

    @property
    def split(self):
        return self.args.split or self.mode

    @property
    def Dataset(self):
        raise NotImplementedError

    @property
    def Optimizer(self):
        return torch.optim.Adam

    def update_args(self, payload, ignored=[]):
        if type(ignored) is str:
            ignored = [ignored]
        ignored += ["__class__"]
        for key in ignored:
            if key in payload:
                del payload[key]
        self.args = getattr(self, "args", argparse.Namespace())
        self.args = argparse.Namespace(**{**vars(self.args), **payload})

    def autofeed(self, f, override={}, mapping={}):
        """Priority: 1. override, 2. parsed args 3. parameters' default"""
        assert hasattr(self, "args")
        payload = vars(self.args)

        def m(key):
            return mapping[key] if key in mapping else key

        params = [p.name for p in inspect.signature(f).parameters.values()]
        kwargs = {k: payload[m(k)] for k in params if m(k) in payload}
        kwargs.update(override)

        return f(**kwargs)

    def create_scheduler(self):
        scheduler = Scheduler()
        if self.mode == "train":
            self.args.lr = scheduler.schedule(self.args.lr)
        return scheduler

    def create_logger(self, epoch=None, label=None):
        """Create a logger to log_dir/name/mode/split/epoch/label,
        Args:
            label: a short name to differentiate different experiments
        Returns:
            logger
        """
        args = self.args
        parts = map(
            lambda s: s.lstrip("/"),
            [
                self.name,
                self.mode,
                self.split,
                "" if epoch is None else str(epoch),
                "default" if label is None else str(label),
            ],
        )
        run_log_dir = Path(args.log_dir, *parts)
        loss_smoothing = [r"(\S+_)?loss(\S+_)?"]
        postfix = run_log_dir.relative_to(Path(args.log_dir, self.name))
        logger = Logger(run_log_dir, loss_smoothing, postfix=postfix)
        return logger

    def create_pbar(self, iterable):
        pbar = tqdm.tqdm(iterable, dynamic_ncols=True, disable=self.args.quiet)

        close = pbar.close

        create_line = lambda: tqdm.tqdm(bar_format="╰{postfix}")

        if self.args.quiet:
            items = {}
            line = create_line()

            timer = Timer(10)

            def set_line(i, s):
                items[i] = s
                if timer.timeup():
                    line.set_postfix_str(", ".join(items.values()) + "\n")
                    timer.restart()

            def close_all():
                close()
                line.close()

        else:
            lines = defaultdict(create_line)

            def set_line(i, s):
                lines[i].set_postfix_str(s)

            def close_all():
                close()
                for line in lines.values():
                    line.close()

        pbar.set_line = set_line
        pbar.close = close_all

        return pbar

    def create_dataset(self):
        raise NotImplementedError

    def create_data_loader(self, **kwargs):
        args = self.args
        dataset = self.create_dataset()
        data_loader = self.autofeed(
            DataLoader,
            override=dict(
                dataset=dataset,
                **kwargs,
            ),
            mapping=dict(
                num_workers="nj",
            ),
        )
        print("Dataset size:", len(dataset))
        return data_loader

    def create_model(self):
        raise NotImplementedError

    def create_optimizer(self, model):
        optimizer = self.autofeed(
            self.Optimizer,
            dict(params=model.parameters(), lr=1),
        )

        for state in optimizer.state.values():
            for k, v in state.items():
                if torch.is_tensor(v):
                    state[k] = v.to(self.args.device)

        def get_lr():
            return optimizer.param_groups[0]["lr"]

        def set_lr(lr):
            for g in optimizer.param_groups:
                g["lr"] = lr

        optimizer.get_lr = get_lr
        optimizer.set_lr = set_lr

        return optimizer

    def initialize_amp(self):
        if self.optimizer is None:
            args = [self.model]
        else:
            args = [self.model, self.optimizer]

        try:
            from apex import amp
        except:
            amp = None

        amp_level = self.args.amp_level
        if amp_level is not None and amp is not None:
            args = list(amp.initialize(*args, opt_level=amp_level))
        else:
            amp = None

        args[0].amp = amp

        self.model = args[0]

        if len(args) > 1:
            self.optimizer = args[1]

    def initialize_saver(self):
        args = self.args
        self.saver = Saver(args.ckpt_dir / self.name)
        if self.training and not self.saver.is_empty and not args.continue_:
            print("Some checkpoints exist and continue not set, skip training.")
            exit()
        self.saver.load(
            self.model,
            self.optimizer,
            epoch=args.epoch,
            strict=args.strict_loading,
        )

    def initialize(self):
        args = self.args

        # scheduler
        self.scheduler = self.create_scheduler()

        # model
        self.model = self.create_model()
        self.model.to(args.device)
        self.model.amp = None

        # optimizer
        if self.training:
            self.optimizer = self.create_optimizer(self.model)

        # amp
        self.initialize_amp()

        # saver
        self.initialize_saver()

        # logger
        if self.training:
            self.data_loader = self.create_data_loader(shuffle=True)
            self.logger = self.create_logger()
        else:
            self.data_loader = self.create_data_loader(shuffle=False)
            self.logger = self.create_logger(self.model.epoch, args.label)

        # events
        if self.training:

            def save(epoch):
                if epoch % args.save_every == 0:
                    self.saver.save(self.model, self.optimizer)

            self.events.epoch_completed.append(save)

    def step(self, batch):
        raise NotImplementedError

    def train_loop(self):
        args = self.args
        model = self.model
        logger = self.logger

        while model.epoch < args.max_epochs:
            model.epoch += 1
            self.events.epoch_started(model.epoch)

            pbar = self.create_pbar(self.data_loader)
            pbar.set_description(f"Train: {model.epoch}/{args.max_epochs}")
            for batch in pbar:
                model.iteration += 1
                self.events.iteration_started(model.iteration)
                self.scheduler.step(model.epoch, model.iteration)
                self.step(batch)
                for i, line in enumerate(logger.render(model.iteration)):
                    pbar.set_line(i, line)
                self.events.iteration_completed(model.iteration)
            pbar.close()

            self.events.epoch_completed(model.epoch)

    @torch.no_grad()
    def validate_loop(self):
        args = self.args
        model = self.model.eval()
        logger = self.logger

        pbar = self.create_pbar(self.data_loader)
        pbar.set_description(f"Validate @{model.epoch}")
        for index, batch in enumerate(pbar):
            self.events.iteration_started(index)
            self.step(batch)
            for i, line in enumerate(logger.render(index)):
                pbar.set_line(i, line)
            self.events.iteration_completed(index)
        mean = lambda l: sum(l) / len(l)
        for key in logger:
            if "loss" in key:
                print(f"Average {key}: {logger.average(key):.4g}")

    @staticmethod
    def try_rmtree(path):
        if path.exists():
            shutil.rmtree(path)
            print(str(path), "removed.")

    @zouqi.command
    def train(
        self,
        lr: str = "1e-3",
        weight_decay: float = 0,
        max_epochs: int = 100,
        save_every: int = 1,
        continue_: flag = False,
        epoch: int = None,
    ):
        self.update_args(locals(), "self")
        self.initialize()
        self.train_loop()

    @zouqi.command
    def validate(self, epoch: int = None, label: str = None):
        self.update_args(locals(), "self")
        self.initialize()
        self.validate_loop()

    @zouqi.command
    def clear(self):
        if input("Are you sure to clear? (y)\n").lower() == "y":
            self.try_rmtree(Path(self.args.ckpt_dir, self.name))
            self.try_rmtree(Path(self.args.log_dir, self.name))
        else:
            print(f"Not cleared.")



import torch
from pathlib import Path


def load_state_dict(module, state_dict, strict):
    if strict:
        module.load_state_dict(state_dict, strict=True)
    else:
        provided = set(state_dict)
        required = set(module.state_dict())
        agreed = provided & required
        state_dict = {k: state_dict[k] for k in agreed}
        print("Provided but not required keys: ")
        print(provided - required)
        print("Required but not provided keys: ")
        print(required - provided)
        module.load_state_dict(state_dict, strict=False)
    return module


class Saver:
    def __init__(self, root):
        self.root = root
        self.cache = {}
        self.last_epoch = self.get_last_epoch()

    @property
    def is_empty(self):
        return self.last_epoch == 0

    def get_last_epoch(self):
        paths = list(self.root.glob("*.pth"))
        if paths:
            last_epoch = int(max(paths, key=lambda p: int(p.stem)).stem)
        else:
            last_epoch = 0
        return last_epoch

    def load(self, model, optimizer=None, epoch=None, strict=True):
        epoch = self.last_epoch if epoch is None else epoch
        iteration = 0

        if epoch > 0:
            path = self.root / f"{epoch}.pth"

            if path not in self.cache:
                self.cache[path] = torch.load(path)

            state_dict = self.cache[path]

            load_state_dict(model, state_dict["model"], strict)

            iteration = state_dict.get("iteration", 0)

            # during testing, optimizer is None
            if optimizer is not None:
                try:
                    optimizer.load_state_dict(state_dict["optimizer"])
                except Exception as e:
                    print(e)
                    # allow not loading the optimizer
                    print("Warning: loading optimizer failed.")

            # without fp16, amp is None
            if model.amp is not None:
                model.amp.load_state_dict(state_dict["amp"])

            print(f"{path} loaded.")

        model.epoch = epoch
        model.iteration = iteration

    def save(self, model, optimizer=None):
        self.root.mkdir(parents=True, exist_ok=True)
        path = self.root / f"{model.epoch}.pth"
        state_dict = dict(
            iteration=model.iteration,
            epoch=model.epoch,
            model=model.state_dict() if model else None,
            optimizer=optimizer.state_dict() if optimizer else None,
            amp=model.amp.state_dict() if model.amp else None,
        )
        torch.save(state_dict, path)
        print(f"{path} saved.")
